package com.raulvibo.tester;

import android.util.Log;

import org.opencv.core.CvType;
import org.opencv.core.Mat;
import org.opencv.core.MatOfPoint;
import org.opencv.core.MatOfPoint2f;
import org.opencv.core.Point;
import org.opencv.core.Rect;
import org.opencv.core.Scalar;
import org.opencv.core.Size;
import org.opencv.imgproc.Imgproc;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

/**
 * Created by raulvibo on 19/07/2017.
 */

public class Procesador {
    MatOfPoint2f approx = new MatOfPoint2f();
    Mat salidaGris, pasoBajo;
    String lectura;
    boolean lecturaEstable;
    int repeticiones;
    Scalar colorRespuesta;
    Scalar colorRespuestaIncorrecta;
    Scalar colorRespuestaCorrecta;
    Scalar colorNumeros;
    Scalar colorBordeCasilla;
    Scalar colorFondoCasilla;
    Scalar colorSuspenso;
    Scalar colorAprobado;
    Scalar colorTexto;

    public Procesador() { //Constructor
        salidaGris = new Mat();
        pasoBajo = new Mat();
        lectura = "";
        repeticiones = 0;
        lecturaEstable = false;
        colorNumeros = new Scalar(77, 40, 0);
        colorBordeCasilla = new Scalar(0, 0, 0);
        colorFondoCasilla = new Scalar(217, 217, 217);
        colorRespuesta = new Scalar(0, 102, 204);
        colorRespuestaCorrecta = new Scalar(102, 153, 0);
        colorRespuestaIncorrecta = new Scalar(204, 0, 0);
        colorSuspenso = new Scalar(204, 0, 0);
        colorAprobado = new Scalar(102, 153, 0);
        colorTexto = new Scalar(0, 102, 204);
    }

    public boolean procesa(Mat entradaGris, Mat salida, Test test) {
        // Binarización
        binarizacionAdaptativa(entradaGris, salidaGris);
        // Filtro morfológico para hacer los objetos oscuros más gruesos.
        erosionar(salidaGris, salidaGris);
        // Segmentar.
        List<Rect> segmentos = segmentar(salidaGris);
        // Ordenar segmentos
        ordenarSegmentos(segmentos);
        // Eliminar segmentos dispares.
        segmentos = eliminarSegmentosDispares(segmentos, areaMasFrecuente(segmentos));
        // Restaurar
        restaurar(entradaGris, segmentos, salidaGris);
        // Rellenar el test
        test.rellenar(entradaGris, segmentos);
        // Convertimos a color
        Imgproc.cvtColor(salidaGris, salida, Imgproc.COLOR_GRAY2RGBA);
        // Comprobamos si es una lectura estable
        lecturaEstable = lecturaEstable(test.getCodificacion());
        return lecturaEstable;
    }

    void erosionar(Mat entrada, Mat salida) {
        double tam = 5;
        Mat SE = Imgproc.getStructuringElement(Imgproc.MORPH_RECT, new Size(tam, tam));
        Imgproc.erode(entrada, salida, SE);
    }

    void binarizacionAdaptativa(Mat entrada, Mat salida) {
        int contraste = 40;
        int tamano = 15;
        Imgproc.adaptiveThreshold(entrada, salida, 255,
                Imgproc.ADAPTIVE_THRESH_MEAN_C,
                Imgproc.THRESH_BINARY,
                tamano, contraste);
    }

    List<Rect> segmentar(Mat entrada) {
        List<Rect> listaRect = new ArrayList<>();
        // A continuación realiza una segmentación eligiendo objetos que cumplan determinadas condiciones.
        List<MatOfPoint> blobs = new ArrayList<>();
        Mat hierarchy = new Mat();
        Imgproc.findContours(entrada, blobs, hierarchy,
                Imgproc.RETR_CCOMP, Imgproc.CHAIN_APPROX_NONE);
        int minimumSize = 15;
        int maximumSize = 60;
        for (int i = 0; blobs != null && i < blobs.size(); i++) {
            MatOfPoint c = blobs.get(i);
            double[] data = hierarchy.get(0, i);
            int parent = (int) data[3];
            int firstChild = (int) data[2];
            if (parent < 0 && firstChild >= 0)
                continue;
            Rect BB = Imgproc.boundingRect(c);
            // Comprobar tamaño
            if (BB.height < minimumSize || BB.width < minimumSize) {
                continue;
            }
            if (BB.height > maximumSize || BB.width > maximumSize) {
                continue;
            }
            // Comprobar anchura similar a altura
            float wf = BB.width;
            float hf = BB.height;
            float ratio = wf / hf;
            if (ratio < 0.75 || ratio > 1.25) {
                continue;
            }
            // Comprobar que tiene cuatro vértices
            if (!tieneCuatroVertices(c)) {
                continue;
            }
            // Comprobar forma cuadrada
            if (!tieneFormaCuadrada(c, BB)) {
                continue;
            }
            // Aqui cumple todos los criterios.
            final Point P1 = new Point(BB.x, BB.y);
            final Point P2 = new Point(BB.x + BB.width, BB.y + BB.height);
            listaRect.add(new Rect((int) P1.x, (int) P1.y, (int) (P2.x - P1.x), (int) (P2.y - P1.y)));
        }
        return listaRect;
    }

    public boolean tieneCuatroVertices(MatOfPoint c) {
        double precisionAproximacion = Imgproc.arcLength(new MatOfPoint2f(c.toArray()), true) * 0.03;
        Imgproc.approxPolyDP(new MatOfPoint2f(c.toArray()), approx, precisionAproximacion, true);
        //Compruebo si tiene cuatro vertices el contorno
        //Las coordenadas de los vertices se almacenan en approx
        return (approx.size().height == 4.0);
    }

    public boolean tieneFormaCuadrada(MatOfPoint c, Rect BB) {
        double umbral = 1;
        Point[] pd = new Point[4];
        pd[0] = new Point(BB.x, BB.y);
        pd[1] = new Point(BB.x + BB.width - 1, BB.y);
        pd[2] = new Point(BB.x + BB.width - 1, BB.y + BB.height - 1);
        pd[3] = new Point(BB.x, BB.y + BB.height - 1);
        MatOfPoint d = new MatOfPoint(pd);
        double distancia = Imgproc.matchShapes(c, d, Imgproc.CV_CONTOURS_MATCH_I1, 0);
        return (distancia <= umbral);
    }

    void restaurar(Mat entradaGris, List<Rect> segmentos, Mat salidaGris) {
        for (int s = 0; s < segmentos.size(); s++) {
            Rect BB = segmentos.get(s);
            entradaGris.submat(BB).copyTo(salidaGris.submat(BB));
        }
    }

    void dibujarCasillas(Mat entrada, Test test) {
        //int numeroCasillas = 0;
        for (int p = 0; p < test.getNumeroPreguntas(); p++) {
            Pregunta pregunta = test.getPregunta(p);
            for (int o = 0; o < pregunta.getNumeroOpciones(); o++) {
                //numeroCasillas++;
                Opcion opcion = pregunta.getOpcion(o);
                Rect casilla = opcion.getCasilla();
                boolean valor = opcion.getValor();
                if (valor) {
                    Imgproc.rectangle(entrada, casilla.tl(), casilla.br(), colorFondoCasilla, -1);
                    dibujarCruz(entrada, casilla, colorRespuesta);
                    Imgproc.rectangle(entrada, casilla.tl(), casilla.br(), colorBordeCasilla, 3);
                } else {
                    Imgproc.rectangle(entrada, casilla.tl(), casilla.br(), colorFondoCasilla, -1);
                    Imgproc.rectangle(entrada, casilla.tl(), casilla.br(), colorBordeCasilla, 3);
                }
                /*Imgproc.putText(entrada, "" + numeroCasillas,
                        casilla.br(), 0, 0.8,
                        colorNumeros, 2, 8, false);*/
            }
        }
    }

    void dibujarCorreccion(Mat entrada, Test test, String correccion) {
        for (int i = 0, p = 0, o = 0; i < correccion.length(); i++) {
            if (correccion.charAt(i) == '*') {
                p++;
                o = 0;
            } else {
                Pregunta pregunta = test.getPregunta(p);
                Opcion opcion = pregunta.getOpcion(o);
                Rect casilla = opcion.getCasilla();
                switch (correccion.charAt(i)) {
                    case 'T':
                        Imgproc.rectangle(entrada, casilla.tl(), casilla.br(), colorFondoCasilla, -1);
                        dibujarTick(entrada, casilla, colorRespuestaCorrecta);
                        Imgproc.rectangle(entrada, casilla.tl(), casilla.br(), colorBordeCasilla, 3);
                        break;
                    case 't':
                        Imgproc.rectangle(entrada, casilla.tl(), casilla.br(), colorFondoCasilla, -1);
                        Imgproc.circle(entrada, new Point(casilla.x + casilla.width / 2, casilla.y + casilla.height / 2), (int) (casilla.height * 0.20), colorRespuestaCorrecta, -1);
                        Imgproc.rectangle(entrada, casilla.tl(), casilla.br(), colorBordeCasilla, 3);
                        break;
                    case 'F':
                        Imgproc.rectangle(entrada, casilla.tl(), casilla.br(), colorFondoCasilla, -1);
                        dibujarCruz(entrada, casilla, colorRespuestaIncorrecta);
                        Imgproc.rectangle(entrada, casilla.tl(), casilla.br(), colorBordeCasilla, 3);
                        break;
                    case 'f':
                        Imgproc.rectangle(entrada, casilla.tl(), casilla.br(), colorFondoCasilla, -1);
                        Imgproc.rectangle(entrada, casilla.tl(), casilla.br(), colorBordeCasilla, 3);
                        break;
                }
                o++;
            }
        }
    }

    public void dibujarCruz(Mat entrada, Rect casilla, Scalar color) {
        Point supIzq = new Point(casilla.x + casilla.width * 0.25, casilla.y + casilla.height * 0.25);
        Point supDer = new Point(casilla.x + casilla.width * 0.75, casilla.y + casilla.height * 0.25);
        Point infDer = new Point(casilla.x + casilla.width * 0.75, casilla.y + casilla.height * 0.75);
        Point infIzq = new Point(casilla.x + casilla.width * 0.25, casilla.y + casilla.height * 0.75);
        Imgproc.line(entrada, supIzq, infDer, color, 3);
        Imgproc.line(entrada, infIzq, supDer, color, 3);
    }

    public void dibujarTick(Mat entrada, Rect casilla, Scalar color) {
        Point izq = new Point(casilla.x + casilla.width * 0.25, casilla.y + casilla.height / 2);
        Point inf = new Point(casilla.x + casilla.width * 0.40, casilla.y + casilla.height * 0.80);
        Point der = new Point(casilla.x + casilla.width * 0.8, casilla.y + casilla.height * 0.20);
        Imgproc.line(entrada, izq, inf, color, 3);
        Imgproc.line(entrada, inf, der, color, 3);
    }

    // https://stackoverflow.com/questions/22131032/detecting-triangles-in-opencv-approxpoly
    public boolean isContourSquare(MatOfPoint thisContour) {
        Rect ret = null;
        MatOfPoint2f thisContour2f = new MatOfPoint2f();
        MatOfPoint approxContour = new MatOfPoint();
        MatOfPoint2f approxContour2f = new MatOfPoint2f();
        thisContour.convertTo(thisContour2f, CvType.CV_32FC2);
        Imgproc.approxPolyDP(thisContour2f, approxContour2f, 2, true);
        approxContour2f.convertTo(approxContour, CvType.CV_32S);
        if (approxContour.size().height == 4) {
            ret = Imgproc.boundingRect(approxContour);
        }
        return (ret != null);
    }

    double areaMasFrecuente(List<Rect> segmentos) {
        double areaMasFrecuente = -1, areaMasFrecuenteA, areaMasFrecuenteB;
        int frecuencia = 0;
        int frecuenciaTmp;
        for (int a = 0; a < segmentos.size(); a++) {
            areaMasFrecuenteA = segmentos.get(a).size().area();
            frecuenciaTmp = 0;
            for (int b = 0; b < segmentos.size(); b++) {
                areaMasFrecuenteB = segmentos.get(b).size().area();
                if (areaMasFrecuenteA == areaMasFrecuenteB) frecuenciaTmp++;
            }
            if (frecuenciaTmp > frecuencia) {
                areaMasFrecuente = areaMasFrecuenteA;
                frecuencia = frecuenciaTmp;
            }
        }
        return areaMasFrecuente;
    }

    public List<Rect> eliminarSegmentosDispares(List<Rect> contours, double areaMasFrecuente) {
        List<Rect> salida = new ArrayList<Rect>();
        double margen = 0.50 * areaMasFrecuente;
        for (int c = 0; c < contours.size(); c++) {
            double area = contours.get(c).size().area();
            if (Math.abs(area - areaMasFrecuente) <= margen) {
                salida.add(contours.get(c));
            }
        }
        return salida;
    }

    // Reference: http://stackoverflow.com/questions/9109890/android-java-how-to-sort-a-list-of-objects-by-a-certain-value-within-the-object
    void ordenarSegmentos(List<Rect> segmentosDisco) {
        try {
            Collections.sort(segmentosDisco, new Comparator<Rect>() {
                public int compare(Rect obj1, Rect obj2) {
                    if (Math.abs(obj1.y - obj2.y) > (obj1.height)) {
                        return (Integer.valueOf(obj1.y).compareTo(obj2.y)); // To compare integer values
                    } else {
                        return (Integer.valueOf(obj1.x).compareTo(obj2.x)); // To compare integer values
                    }
                }
            });
        }catch(Exception e){
            // Si la comparación produce una violación de su contrato general, es mejor eliminar los segmentos.
            Log.e("Tester", e.toString());
            segmentosDisco.clear();
        }
    }

    boolean lecturaEstable(String lectura) {
        if (this.lectura.equals(lectura) && lectura.length() > 2) {
            repeticiones++;
        } else {
            this.lectura = lectura;
            repeticiones = 1;
        }
        return repeticiones > 2;
    }

    public void mostrarPuntuacion(Mat entrada, double puntuacion, double puntuacionMaxima) {
        String mensaje = String.format("%.2f", puntuacion);
        Point centro = new Point(entrada.width() * 0.94, entrada.height() * 0.105);
        Point punto;
        int radio;
        if (puntuacion < 10) {
            radio = (int) Math.round(entrada.height() * 0.08);
            punto = new Point(entrada.width() * 0.906, entrada.height() * 0.12);
        } else {
            radio = (int) Math.round(entrada.height() * 0.09);
            punto = new Point(entrada.width() * 0.897, entrada.height() * 0.12);
        }
        if (puntuacion < (puntuacionMaxima / 2)) {
            //Imgproc.circle(entrada, centro, radio, colorSuspenso, -1);
            Imgproc.circle(entrada, centro, radio, colorBordeCasilla, 10);
            Imgproc.circle(entrada, centro, radio, colorSuspenso, 5);
            mostrarMensaje(entrada, punto, mensaje, 1.2, colorBordeCasilla, colorSuspenso);
        } else {
            //Imgproc.circle(entrada, centro, radio, colorAprobado, -1);
            Imgproc.circle(entrada, centro, radio, colorBordeCasilla, 10);
            Imgproc.circle(entrada, centro, radio, colorAprobado, 5);
            mostrarMensaje(entrada, punto, mensaje, 1.2, colorBordeCasilla, colorAprobado);
        }
    }

    public void mostrarMensaje(Mat entrada, Point punto, String mensaje, double fontScale, Scalar colorBorde, Scalar colorRelleno) {
        int fontFace = 4;
        int thickness = 5;
        Imgproc.putText(entrada, mensaje,
                punto, fontFace, fontScale,
                colorBorde, thickness, 8, false);
        Imgproc.putText(entrada, mensaje,
                punto, fontFace, fontScale,
                colorRelleno, thickness / 2, 8, false);
    }

    public void mostrarMensajeTexto(Mat entrada, Point punto, String mensaje, double fontScale, Scalar color) {
        int fontFace = 1;
        int thickness = 2;
        Imgproc.putText(entrada, mensaje,
                punto, fontFace, fontScale,
                colorBordeCasilla, thickness, 8, false);
        Imgproc.putText(entrada, mensaje,
                punto, fontFace, fontScale,
                color, thickness / 2, 8, false);
    }

    public void mostrarTitulo(Mat entrada, String titulo, String usuario, String fecha) {
        Point point = new Point(entrada.width() * 0.01, entrada.height() * 0.04);
        mostrarMensajeTexto(entrada, point, titulo, 2, colorTexto);
        Point punto = new Point(entrada.width() * 0.695, entrada.height() * 0.935);
        mostrarMensajeTexto(entrada, punto, usuario, 2, colorTexto);
        punto = new Point(entrada.width() * 0.695, entrada.height() * 0.985);
        mostrarMensajeTexto(entrada, punto, fecha, 2, colorTexto);
    }
}

